@@ -1,16 +1,20 @@
+"""Mapper and Sub-Mapper"""
 import re
 import sys
 import threading
-import threadinglocal
 
-if sys.version < '2.4':
-    from sets import ImmutableSet as frozenset
+import pkg_resources
+from repoze.lru import LRUCache
 
 from routes import request_config
 from routes.util import controller_scan, MatchException, RoutesException
 from routes.route import Route
 
 
+COLLECTION_ACTIONS = ['index', 'create', 'new']
+MEMBER_ACTIONS = ['show', 'update', 'delete', 'edit']
+
+
 def strip_slashes(name):
     """Remove slashes from the beginning and end of a part/URL."""
     if name.startswith('/'):
@@ -20,7 +24,262 @@ def strip_slashes(name):
     return name
 
 
-class Mapper(object):
+class SubMapperParent(object):
+    """Base class for Mapper and SubMapper, both of which may be the parent
+    of SubMapper objects
+    """
+    
+    def submapper(self, **kargs):
+        """Create a partial version of the Mapper with the designated
+        options set
+        
+        This results in a :class:`routes.mapper.SubMapper` object.
+        
+        If keyword arguments provided to this method also exist in the
+        keyword arguments provided to the submapper, their values will
+        be merged with the saved options going first.
+        
+        In addition to :class:`routes.route.Route` arguments, submapper
+        can also take a ``path_prefix`` argument which will be
+        prepended to the path of all routes that are connected.
+        
+        Example::
+            
+            >>> map = Mapper(controller_scan=None)
+            >>> map.connect('home', '/', controller='home', action='splash')
+            >>> map.matchlist[0].name == 'home'
+            True
+            >>> m = map.submapper(controller='home')
+            >>> m.connect('index', '/index', action='index')
+            >>> map.matchlist[1].name == 'index'
+            True
+            >>> map.matchlist[1].defaults['controller'] == 'home'
+            True
+        
+        Optional ``collection_name`` and ``resource_name`` arguments are
+        used in the generation of route names by the ``action`` and
+        ``link`` methods.  These in turn are used by the ``index``,
+        ``new``, ``create``, ``show``, ``edit``, ``update`` and
+        ``delete`` methods which may be invoked indirectly by listing
+        them in the ``actions`` argument.  If the ``formatted`` argument
+        is set to ``True`` (the default), generated paths are given the
+        suffix '{.format}' which matches or generates an optional format
+        extension.
+        
+        Example::
+        
+            >>> from routes.util import url_for
+            >>> map = Mapper(controller_scan=None)
+            >>> m = map.submapper(path_prefix='/entries', collection_name='entries', resource_name='entry', actions=['index', 'new'])
+            >>> url_for('entries') == '/entries'
+            True
+            >>> url_for('new_entry', format='xml') == '/entries/new.xml'
+            True
+
+        """
+        return SubMapper(self, **kargs)
+
+    def collection(self, collection_name, resource_name, path_prefix=None,
+                   member_prefix='/{id}', controller=None,
+                   collection_actions=COLLECTION_ACTIONS,
+                   member_actions = MEMBER_ACTIONS, member_options=None,
+                   **kwargs):
+        """Create a submapper that represents a collection.
+
+        This results in a :class:`routes.mapper.SubMapper` object, with a
+        ``member`` property of the same type that represents the collection's
+        member resources.
+        
+        Its interface is the same as the ``submapper`` together with
+        ``member_prefix``, ``member_actions`` and ``member_options``
+        which are passed to the ``member` submatter as ``path_prefix``,
+        ``actions`` and keyword arguments respectively.
+        
+        Example::
+        
+            >>> from routes.util import url_for
+            >>> map = Mapper(controller_scan=None)
+            >>> c = map.collection('entries', 'entry')
+            >>> c.member.link('ping', method='POST')
+            >>> url_for('entries') == '/entries'
+            True
+            >>> url_for('edit_entry', id=1) == '/entries/1/edit'
+            True
+            >>> url_for('ping_entry', id=1) == '/entries/1/ping'
+            True
+
+        """
+        if controller is None:
+            controller = resource_name or collection_name
+        
+        if path_prefix is None:
+            path_prefix = '/' + collection_name
+
+        collection = SubMapper(self, collection_name=collection_name,
+                               resource_name=resource_name,
+                               path_prefix=path_prefix, controller=controller,
+                               actions=collection_actions, **kwargs)
+        
+        collection.member = SubMapper(collection, path_prefix=member_prefix,
+                                      actions=member_actions, 
+                                      **(member_options or {}))
+
+        return collection
+
+
+class SubMapper(SubMapperParent):
+    """Partial mapper for use with_options"""
+    def __init__(self, obj, resource_name=None, collection_name=None,
+                 actions=None, formatted=None, **kwargs):
+        self.kwargs = kwargs
+        self.obj = obj
+        self.collection_name = collection_name
+        self.member = None
+        self.resource_name = resource_name \
+                            or getattr(obj, 'resource_name', None) \
+                            or kwargs.get('controller', None) \
+                            or getattr(obj, 'controller', None)
+        if formatted is not None:
+            self.formatted = formatted
+        else:
+            self.formatted = getattr(obj, 'formatted', None)
+            if self.formatted is None:
+                self.formatted = True
+
+        self.add_actions(actions or [])
+        
+    def connect(self, *args, **kwargs):
+        newkargs = {}
+        newargs = args
+        for key, value in self.kwargs.items():
+            if key == 'path_prefix':
+                if len(args) > 1:
+                    newargs = (args[0], self.kwargs[key] + args[1])
+                else:
+                    newargs = (self.kwargs[key] + args[0],)
+            elif key in kwargs:
+                if isinstance(value, dict):
+                    newkargs[key] = dict(value, **kwargs[key]) # merge dicts
+                else:
+                    newkargs[key] = value + kwargs[key]
+            else:
+                newkargs[key] = self.kwargs[key]
+        for key in kwargs:
+            if key not in self.kwargs:
+                newkargs[key] = kwargs[key]
+        return self.obj.connect(*newargs, **newkargs)
+
+    def link(self, rel=None, name=None, action=None, method='GET',
+             formatted=None, **kwargs):
+        """Generates a named route for a subresource.
+
+        Example::
+        
+            >>> from routes.util import url_for
+            >>> map = Mapper(controller_scan=None)
+            >>> c = map.collection('entries', 'entry')
+            >>> c.link('recent', name='recent_entries')
+            >>> c.member.link('ping', method='POST', formatted=True)
+            >>> url_for('entries') == '/entries'
+            True
+            >>> url_for('recent_entries') == '/entries/recent'
+            True
+            >>> url_for('ping_entry', id=1) == '/entries/1/ping'
+            True
+            >>> url_for('ping_entry', id=1, format='xml') == '/entries/1/ping.xml'
+            True
+
+        """
+        if formatted or (formatted is None and self.formatted):
+            suffix = '{.format}'
+        else:
+            suffix = ''
+
+        return self.connect(name or (rel + '_' + self.resource_name),
+                            '/' + (rel or name) + suffix,
+                            action=action or rel or name,
+                            **_kwargs_with_conditions(kwargs, method))
+
+    def new(self, **kwargs):
+        """Generates the "new" link for a collection submapper."""
+        return self.link(rel='new', **kwargs)
+
+    def edit(self, **kwargs):
+        """Generates the "edit" link for a collection member submapper."""
+        return self.link(rel='edit', **kwargs)
+
+    def action(self, name=None, action=None, method='GET', formatted=None,
+               **kwargs):
+        """Generates a named route at the base path of a submapper.
+
+        Example::
+        
+            >>> from routes import url_for
+            >>> map = Mapper(controller_scan=None)
+            >>> c = map.submapper(path_prefix='/entries', controller='entry')
+            >>> c.action(action='index', name='entries', formatted=True)
+            >>> c.action(action='create', method='POST')
+            >>> url_for(controller='entry', action='index', method='GET') == '/entries'
+            True
+            >>> url_for(controller='entry', action='index', method='GET', format='xml') == '/entries.xml'
+            True
+            >>> url_for(controller='entry', action='create', method='POST') == '/entries'
+            True
+
+        """
+        if formatted or (formatted is None and self.formatted):
+            suffix = '{.format}'
+        else:
+            suffix = ''
+        return self.connect(name or (action + '_' + self.resource_name),
+                            suffix,
+                            action=action or name,
+                            **_kwargs_with_conditions(kwargs, method))
+            
+    def index(self, name=None, **kwargs):
+        """Generates the "index" action for a collection submapper."""
+        return self.action(name=name or self.collection_name,
+                           action='index', method='GET', **kwargs)
+
+    def show(self, name = None, **kwargs):
+        """Generates the "show" action for a collection member submapper."""
+        return self.action(name=name or self.resource_name,
+                           action='show', method='GET', **kwargs)
+
+    def create(self, **kwargs):
+        """Generates the "create" action for a collection submapper."""
+        return self.action(action='create', method='POST', **kwargs)
+        
+    def update(self, **kwargs):
+        """Generates the "update" action for a collection member submapper."""
+        return self.action(action='update', method='PUT', **kwargs)
+
+    def delete(self, **kwargs):
+        """Generates the "delete" action for a collection member submapper."""
+        return self.action(action='delete', method='DELETE', **kwargs)
+
+    def add_actions(self, actions):
+        [getattr(self, action)() for action in actions]
+
+    # Provided for those who prefer using the 'with' syntax in Python 2.5+
+    def __enter__(self):
+        return self
+    
+    def __exit__(self, type, value, tb):
+        pass
+
+# Create kwargs with a 'conditions' member generated for the given method
+def _kwargs_with_conditions(kwargs, method):
+    if method and 'conditions' not in kwargs:
+        newkwargs = kwargs.copy()
+        newkwargs['conditions'] = {'method': method}                
+        return newkwargs             
+    else:
+        return kwargs
+
+
+
+class Mapper(SubMapperParent):
     """Mapper handles URL generation and URL recognition in a web
     application.
     
@@ -33,7 +292,7 @@ class Mapper(object):
     
     """
     def __init__(self, controller_scan=controller_scan, directory=None, 
-                 always_scan=False, register=True, explicit=False):
+                 always_scan=False, register=True, explicit=True):
         """Create a new Mapper instance
         
         All keyword arguments are optional.
@@ -44,6 +303,10 @@ class Mapper(object):
             ``directory`` keyword arg is present, it will be passed
             into the function during its call. This option defaults to
             a function that will scan a directory for controllers.
+            
+            Alternatively, a list of controllers or None can be passed
+            in which are assumed to be the definitive list of
+            controller names valid when matching 'controller'.
         
         ``directory``
             Passed into controller_scan for the directory to scan. It
@@ -98,11 +361,12 @@ class Mapper(object):
         self.matchlist = []
         self.maxkeys = {}
         self.minkeys = {}
-        self.urlcache = {}
+        self.urlcache = LRUCache(1600)
         self._created_regs = False
         self._created_gens = False
+        self._master_regexp = None
         self.prefix = None
-        self.req_data = threadinglocal.local()
+        self.req_data = threading.local()
         self.directory = directory
         self.always_scan = always_scan
         self.controller_scan = controller_scan
@@ -117,12 +381,33 @@ class Mapper(object):
         self.encoding = 'utf-8'
         self.decode_errors = 'ignore'
         self.hardcode_names = True
-        self.minimization = True
+        self.minimization = False
         self.create_regs_lock = threading.Lock()
         if register:
             config = request_config()
             config.mapper = self
     
+    def __str__(self):
+        """Generates a tabular string representation."""
+        def format_methods(r):
+            if r.conditions:
+                method = r.conditions.get('method', '')
+                return type(method) is str and method or ', '.join(method)
+            else:
+                return ''
+
+        table = [('Route name', 'Methods', 'Path')] + \
+                [(r.name or '', format_methods(r), r.routepath or '')
+                 for r in self.matchlist]
+            
+        widths = [max(len(row[col]) for row in table)
+                  for col in range(len(table[0]))]
+        
+        return '\n'.join(
+            ' '.join(row[col].ljust(widths[col])
+                     for col in range(len(widths)))
+            for row in table)
+
     def _envget(self):
         try:
             return self.req_data.environ
@@ -133,7 +418,46 @@ class Mapper(object):
     def _envdel(self):
         del self.req_data.environ
     environ = property(_envget, _envset, _envdel)
-
+    
+    def extend(self, routes, path_prefix=''):
+        """Extends the mapper routes with a list of Route objects
+        
+        If a path_prefix is provided, all the routes will have their
+        path prepended with the path_prefix.
+        
+        Example::
+            
+            >>> map = Mapper(controller_scan=None)
+            >>> map.connect('home', '/', controller='home', action='splash')
+            >>> map.matchlist[0].name == 'home'
+            True
+            >>> routes = [Route('index', '/index.htm', controller='home',
+            ...                 action='index')]
+            >>> map.extend(routes)
+            >>> len(map.matchlist) == 2
+            True
+            >>> map.extend(routes, path_prefix='/subapp')
+            >>> len(map.matchlist) == 3
+            True
+            >>> map.matchlist[2].routepath == '/subapp/index.htm'
+            True
+        
+        .. note::
+            
+            This function does not merely extend the mapper with the
+            given list of routes, it actually creates new routes with
+            identical calling arguments.
+        
+        """
+        for route in routes:
+            if path_prefix and route.minimization:
+                routepath = '/'.join([path_prefix, route.routepath])
+            elif path_prefix:
+                routepath = path_prefix + route.routepath
+            else:
+                routepath = route.routepath
+            self.connect(route.name, routepath, **route._kargs)
+                
     def connect(self, *args, **kargs):
         """Create and connect a new Route to the Mapper.
         
@@ -154,7 +478,8 @@ class Mapper(object):
         routename = None
         if len(args) > 1:
             routename = args[0]
-            args = args[1:]
+        else:
+            args = (None,) + args
         if '_explicit' not in kargs:
             kargs['_explicit'] = self.explicit
         if '_minimize' not in kargs:
@@ -243,20 +568,36 @@ class Mapper(object):
         if clist is None:
             if self.directory:
                 clist = self.controller_scan(self.directory)
-            else:
+            elif callable(self.controller_scan):
                 clist = self.controller_scan()
-            
+            elif not self.controller_scan:
+                clist = []
+            else:
+                clist = self.controller_scan
+        
         for key, val in self.maxkeys.iteritems():
             for route in val:
                 route.makeregexp(clist)
         
+        regexps = []
+        routematches = []
+        for route in self.matchlist:
+            if not route.static:
+                routematches.append(route)
+                regexps.append(route.makeregexp(clist, include_names=False))
+        self._routematches = routematches
         
         # Create our regexp to strip the prefix
         if self.prefix:
             self._regprefix = re.compile(self.prefix + '(.*)')
+        
+        # Save the master regexp
+        regexp = '|'.join(['(?:%s)' % x for x in regexps])
+        self._master_reg = regexp
+        self._master_regexp = re.compile(regexp)
         self._created_regs = True
     
-    def _match(self, url):
+    def _match(self, url, environ):
         """Internal Route matcher
         
         Matches a URL against a route, and returns a tuple of the match
@@ -283,11 +624,19 @@ class Mapper(object):
                     url = '/'
             else:
                 return (None, None, matchlog)
-        environ = self.environ
+                
+        environ = environ or self.environ
         sub_domains = self.sub_domains
         sub_domains_ignore = self.sub_domains_ignore
         domain_match = self.domain_match
         debug = self.debug
+        
+        # Check to see if its a valid url against the main regexp
+        # Done for faster invalid URL elimination
+        valid_url = re.match(self._master_regexp, url)
+        if not valid_url:
+            return (None, None, matchlog)
+        
         for route in self.matchlist:
             if route.static:
                 if debug:
@@ -297,11 +646,11 @@ class Mapper(object):
                                 domain_match)
             if debug:
                 matchlog.append(dict(route=route, regexp=bool(match)))
-            if match:
+            if isinstance(match, dict) or match:
                 return (match, route, matchlog)
         return (None, None, matchlog)
     
-    def match(self, url):
+    def match(self, url=None, environ=None):
         """Match a URL against against one of the routes contained.
         
         Will return None if no valid match is found.
@@ -311,18 +660,20 @@ class Mapper(object):
             resultdict = m.match('/joe/sixpack')
         
         """
-        if not url:
-            raise RoutesException('No URL provided, the minimum URL necessary'
-                                 ' to match is "/".')
+        if not url and not environ:
+            raise RoutesException('URL or environ must be provided')
         
-        result = self._match(url)
+        if not url:
+            url = environ['PATH_INFO']
+                
+        result = self._match(url, environ)
         if self.debug:
             return result[0], result[1], result[2]
-        if result[0]:
+        if isinstance(result[0], dict) or result[0]:
             return result[0]
         return None
     
-    def routematch(self, url):
+    def routematch(self, url=None, environ=None):
         """Match a URL against against one of the routes contained.
         
         Will return None if no valid match is found, otherwise a
@@ -333,10 +684,15 @@ class Mapper(object):
             resultdict, route_obj = m.match('/joe/sixpack')
         
         """
-        result = self._match(url)
+        if not url and not environ:
+            raise RoutesException('URL or environ must be provided')
+        
+        if not url:
+            url = environ['PATH_INFO']
+        result = self._match(url, environ)
         if self.debug:
             return result[0], result[1], result[2]
-        if result[0]:
+        if isinstance(result[0], dict) or result[0]:
             return result[0], result[1]
         return None
     
@@ -363,6 +719,7 @@ class Mapper(object):
             if 'action' not in kargs:
                 kargs['action'] = 'index'
         
+        environ = kargs.pop('_environ', self.environ)
         controller = kargs.get('controller', None)
         action = kargs.get('action', None)
 
@@ -375,14 +732,15 @@ class Mapper(object):
         if self.urlcache is not None:
             if self.environ:
                 cache_key_script_name = '%s:%s' % (
-                    self.environ.get('SCRIPT_NAME', ''), cache_key)
+                    environ.get('SCRIPT_NAME', ''), cache_key)
             else:
                 cache_key_script_name = cache_key
         
             # Check the url cache to see if it exists, use it if it does
             for key in [cache_key, cache_key_script_name]:
-                if key in self.urlcache:
-                    return self.urlcache[key]
+                val = self.urlcache.get(key, self)
+                if val != self:
+                    return val
         
         actionlist = self._gendict.get(controller) or self._gendict.get('*', {})
         if not actionlist and not args:
@@ -404,7 +762,7 @@ class Mapper(object):
             cacheset = True
             newlist = []
             for route in keylist:
-                if len(route.minkeys-keys) == 0:
+                if len(route.minkeys - route.dotkeys - keys) == 0:
                     newlist.append(route)
             keylist = newlist
             
@@ -447,7 +805,7 @@ class Mapper(object):
             keylist.sort(keysort)
             if cacheset:
                 sortcache[cachekey] = keylist
-        
+                
         # Iterate through the keylist of sorted routes (or a single route if
         # it was passed in explicitly for hardcoded named routes)
         for route in keylist:
@@ -456,7 +814,11 @@ class Mapper(object):
                 kval = kargs.get(key)
                 if not kval:
                     continue
-                if kval != route.defaults[key]:
+                if isinstance(kval, str):
+                    kval = kval.decode(self.encoding)
+                else:
+                    kval = unicode(kval)
+                if kval != route.defaults[key] and not callable(route.defaults[key]):
                     fail = True
                     break
             if fail:
@@ -466,14 +828,14 @@ class Mapper(object):
                 if self.prefix:
                     path = self.prefix + path
                 external_static = route.static and route.external
-                if self.environ and self.environ.get('SCRIPT_NAME', '') != ''\
+                if environ and environ.get('SCRIPT_NAME', '') != ''\
                     and not route.absolute and not external_static:
-                    path = self.environ['SCRIPT_NAME'] + path
+                    path = environ['SCRIPT_NAME'] + path
                     key = cache_key_script_name
                 else:
                     key = cache_key
                 if self.urlcache is not None:
-                    self.urlcache[key] = str(path)
+                    self.urlcache.put(key, str(path))
                 return str(path)
             else:
                 continue
@@ -670,6 +1032,7 @@ class Mapper(object):
             '_member_name': member_name,
             '_collection_name': collection_name,
             '_parent_resource': parent_resource,
+            '_filter': kwargs.get('_filter')
         }
         
         def requirements_for(meth):
@@ -719,8 +1082,8 @@ class Mapper(object):
                 self.connect("formatted_" + name_prefix + name, formatted_path, 
                              **route_options)
                 self.connect(name_prefix + name, path, **route_options)
-        
-        requirements_regexp = '[^\/]+'
+
+        requirements_regexp = '[^\/]+(?<!\\\)'
 
         # Add the routes that deal with member methods of a resource
         for method, lst in member_methods.iteritems():
@@ -785,7 +1148,7 @@ class Mapper(object):
         # Create the dict of args for the generation route
         for key in both_args + gen_args:
             if key in kwargs:
-                gen_dict = kwargs[key]
+                gen_dict[key] = kwargs[key]
         gen_dict['_static'] = True
         
         # Create the dict of args for the matching route
